import io
import logging
import textwrap

from rstcloth.cloth import Cloth


logger = logging.getLogger("rstcloth")

# TODO REVIEW THIS ENTIRE FILE - ADDED BY A THIRD PARTY CONTRIBUTOR BUT FOR SOME REASON IT REDEFINES RSTCLOTH


def fill(string, first=0, hanging=0, wrap=True, width=72):
    """

    :param string:
    :param first:
    :param hanging:
    :param wrap:
    :param width:
    :return:
    """
    first_indent = " " * first
    hanging_indent = " " * hanging

    if wrap is True:
        return textwrap.fill(
            string,
            width=width,
            break_on_hyphens=False,
            break_long_words=False,
            initial_indent=first_indent,
            subsequent_indent=hanging_indent,
        )
    else:
        content = string.split("\n")
        if first == hanging:
            return "\n".join([first_indent + line for line in content])
        elif first > hanging:
            indent_diff = first - hanging
            o = indent_diff * " "
            o += "\n".join([hanging_indent + line for line in content])
            return o
        elif first < hanging:
            indent_diff = hanging - first
            o = "\n".join([hanging_indent + line for line in content])
            return o[indent_diff:]


def _indent(content, indent):
    """

    :param content:
    :param indent:
    :return:
    """
    if indent == 0:
        return content
    else:
        indent = " " * indent
        if isinstance(content, list):
            return ["".join([indent, line]) for line in content]
        else:
            return "".join([indent, content])


class RstCloth(Cloth):
    """
    RstCloth is the base object to create a ReStructuredText document programatically.

    :param line_width: (optional, default=72), the line width to use if wrap is set to true in an individual action.
    :return:
    """

    def __init__(self, line_width=72):
        self._line_width = line_width
        self._data = []

    def _add(self, content):
        """

        :param content: the text to write into this element
        :return:
        """

        if isinstance(content, list):
            self._data.extend(content)
        else:
            self._data.append(content)

    def newline(self, count=1):
        """

        :param count: (optional default=1) the number of newlines to add
        :return:
        """

        if isinstance(count, int):
            if count == 1:
                self._add("")
            else:
                # subtract one because every item gets one \n for free.
                self._add("\n" * (count - 1))
        else:
            raise Exception("Count of newlines must be a positive int.")

    def table(self, header, data, indent=0):
        """

        :param header: a list of header values (strings), to use for the table
        :param data: a list of lists of row data (same length as the header list each)
        :param indent: something!
        :return:
        """

        t = Table(header, data=data)
        self._add(_indent("\n" + t.render(), indent))

    def directive(self, name, arg=None, fields=None, content=None, indent=0, wrap=True):
        """

        :param name: the directive itself to use
        :param arg: the argument to pass into the directive
        :param fields: fields to append as children underneath the directive
        :param content: the text to write into this element
        :param indent: (optional default=0) number of characters to indent this element
        :param wrap: (optional, default=True) Whether or not to wrap lines to the line_width
        :return:
        """
        logger.debug("Ignoring wrap parameter, presumably for api consistency. wrap=%s", wrap)
        o = list()
        o.append(".. {0}::".format(name))

        if arg is not None:
            o[0] += " " + arg

        if fields is not None:
            for k, v in fields:
                o.append(_indent(":" + k + ": " + str(v), 3))

        if content is not None:
            o.append("")

            if isinstance(content, list):
                o.extend(_indent(content, 3))
            else:
                o.append(_indent(content, 3))

        self._add(_indent(o, indent))

    @staticmethod
    def role(name, value, text=None):
        """

        :param name: the name of the role
        :param value: the value of the role
        :param text: (optional, default=None) text after the role
        :return:
        """

        if isinstance(name, list):
            name = ":".join(name)

        if text is None:
            return ":{0}:`{1}`".format(name, value)
        else:
            return ":{0}:`{2} <{1}>`".format(name, value, text)

    @staticmethod
    def bold(string):
        """

        :param string: the text to write into this element
        :return:
        """
        return "**{0}**".format(string)

    @staticmethod
    def emph(string):
        """

        :param string: the text to write into this element
        :return:
        """
        return "*{0}*".format(string)

    @staticmethod
    def pre(string):
        """

        :param string: the text to write into this element
        :return:
        """
        return "``{0}``".format(string)

    @staticmethod
    def inline_link(text, link):
        """

        :param text: the printed value of the link
        :param link: the url the link should goto
        :return:
        """
        return "`{0} <{1}>`_".format(text, link)

    @staticmethod
    def footnote_ref(name):
        """

        :param name: the text to write into this element
        :return:
        """
        return "[#{0}]".format(name)

    def _paragraph(self, content, wrap=True):
        """

        :param content: the text to write into this element
        :param wrap: (optional, default=True) Whether or not to wrap lines to the line_width
        :return:
        """
        return [i.rstrip() for i in fill(content, wrap=wrap, width=self._line_width).split("\n")]

    def replacement(self, name, value, indent=0):
        """

        :param name: the name of the replacement
        :param value: the value fo the replacement
        :param indent: (optional default=0) number of characters to indent this element
        :return:
        """

        output = ".. |{0}| replace:: {1}".format(name, value)
        self._add(_indent(output, indent))

    def codeblock(self, content, indent=0, wrap=True, language=None):
        """

        :param content: the text to write into this element
        :param indent: (optional default=0) number of characters to indent this element
        :param wrap: (optional, default=True) Whether or not to wrap lines to the line_width
        :param language:
        :return:
        """
        if language is None:
            o = ["::", _indent(content, 3)]
            self._add(_indent(o, indent))
        else:
            self.directive(name="code-block", arg=language, content=content, indent=indent)

    def footnote(self, ref, text, indent=0, wrap=True):
        """

        :param ref: the reference value
        :param text: the text to write into this element
        :param indent: (optional default=0) number of characters to indent this element
        :param wrap: (optional, default=True) Whether or not to wrap lines to the line_width
        :return:
        """
        self._add(
            fill(
                ".. [#{0}] {1}".format(ref, text),
                indent,
                indent + 3,
                wrap,
                width=self._line_width,
            )
        )

    def definition(self, name, text, indent=0, wrap=True, bold=False):
        """

        :param name: the name of the definition
        :param text: the text to write into this element
        :param indent: (optional default=0) number of characters to indent this element
        :param wrap: (optional, default=True) Whether or not to wrap lines to the line_width
        :param bold:
        :return:
        """
        o = []

        if bold is True:
            name = self.bold(name)

        o.append(_indent(name, indent))
        o.append(fill(text, indent + 3, indent + 3, wrap=wrap, width=self._line_width))

        self._add(o)

    def li(self, content, bullet="-", indent=0, wrap=True):
        """

        :param content: the text to write into this element
        :param bullet: (optional, default='-') the character of the bullet
        :param indent: (optional default=0) number of characters to indent this element
        :param wrap: (optional, default=True) Whether or not to wrap lines to the line_width
        :return:
        """

        bullet += " "
        hanging_indent_len = indent + len(bullet)

        if isinstance(content, list):
            content = bullet + "\n".join(content)
            self._add(
                fill(
                    content,
                    indent,
                    indent + hanging_indent_len,
                    wrap,
                    width=self._line_width,
                )
            )
        else:
            content = bullet + fill(content, 0, len(bullet), wrap, width=self._line_width)
            self._add(fill(content, indent, indent, wrap, width=self._line_width))

    def field(self, name, value, indent=0, wrap=True):
        """

        :param name: the name of the field
        :param value: the value of the field
        :param indent: (optional default=0) number of characters to indent this element
        :param wrap: (optional, default=True) Whether or not to wrap lines to the line_width
        :return:
        """
        output = [":{0}:".format(name)]

        if len(name) + len(value) < 60:
            output[0] += " " + value
            final = True
        else:
            output.append("")
            final = False

        if wrap is True and final is False:
            content = fill(value, wrap=wrap, width=self._line_width).split("\n")
            for line in content:
                output.append(_indent(line, 3))

        if wrap is False and final is False:
            output.append(_indent(value, 3))

        for line in output:
            self._add(_indent(line, indent))

    def ref_target(self, name, indent=0):
        """

        :param name: the name of the reference target
        :param indent: (optional default=0) number of characters to indent this element
        :return:
        """
        o = ".. _{0}:".format(name)
        self._add(_indent(o, indent))

    def content(self, content, indent=0, wrap=True):
        """

        :param content: the text to write into this element
        :param indent: (optional default=0) number of characters to indent this element
        :param wrap: (optional, default=True) Whether or not to wrap lines to the line_width
        :return:
        """
        if isinstance(content, list):
            for line in content:
                self._add(_indent(line, indent))
        else:
            lines = self._paragraph(content, wrap)

            for line in lines:
                self._add(_indent(line, indent))

    def title(self, text, char="=", indent=0):
        """

        :param text: the text to write into this element
        :param char: (optional, default='=') the character to underline the title with
        :param indent: (optional default=0) number of characters to indent this element
        :return:
        """
        line = char * len(text)
        self._add(_indent([line, text, line], indent))

    def heading(self, text, char, indent=0):
        """

        :param text: the text to write into this element
        :param char: the character to line the heading with
        :param indent: (optional default=0) number of characters to indent this element
        :return:
        """
        self._add(_indent([text, char * len(text)], indent))

    def h1(self, text, indent=0):
        """

        :param text: the text to write into this element
        :param indent: (optional default=0) number of characters to indent this element
        :return:
        """
        self.heading(text, char="=", indent=indent)

    def h2(self, text, indent=0):
        """

        :param text: the text to write into this element
        :param indent: (optional default=0) number of characters to indent this element
        :return:
        """
        self.heading(text, char="-", indent=indent)

    def h3(self, text, indent=0):
        """

        :param text: the text to write into this element
        :param indent: (optional default=0) number of characters to indent this element
        :return:
        """
        self.heading(text, char="~", indent=indent)

    def h4(self, text, indent=0):
        """

        :param text: the text to write into this element
        :param indent: (optional default=0) number of characters to indent this element
        :return:
        """
        self.heading(text, char="+", indent=indent)

    def h5(self, text, indent=0):
        """

        :param text: the text to write into this element
        :param indent: (optional default=0) number of characters to indent this element
        :return:
        """
        self.heading(text, char="^", indent=indent)

    def h6(self, text, indent=0):
        """

        :param text: the text to write into this element
        :param indent: (optional default=0) number of characters to indent this element
        :return:
        """
        self.heading(text, char=";", indent=indent)


class Table(object):
    def __init__(self, header, data=None):
        """

        :param header: a list of header values
        :param data: optional, a list of lists of data to add as rows.
        :return:
        """

        self.num_columns = len(header)
        self.num_rows = 0
        self.header = header
        self.rows = []
        if data is not None:
            for row in data:
                self.append(row)

    def append(self, row):
        """

        :param row: a single row to add (list)
        :return:
        """
        row = [str(x) for x in row]

        if len(row) != self.num_columns:
            raise ValueError("row length mismatch")

        self.num_rows += 1
        self.rows.append(row)

        return self

    def _max_col_with(self, idx):
        """

        :param idx: the index to return max width of
        :return:
        """
        return max([len(self.header[idx])] + [len(x[idx]) for x in self.rows])

    def render(self, padding=3):
        """

        :return:
        """
        widths = [self._max_col_with(x) + padding for x in range(self.num_columns)]
        f = io.StringIO()

        # first right out the header
        f.write("+")
        for width in widths:
            f.write("-" * width + "+")
        f.write("\n")

        f.write("|")
        for col, width in zip(self.header, widths):
            f.write(col + " " * (width - len(col)) + "|")
        f.write("\n")

        f.write("+")
        for width in widths:
            f.write("=" * width + "+")
        f.write("\n")

        # then the rows:
        for ridx in range(self.num_rows):
            f.write("|")
            for col, width in zip(self.rows[ridx], widths):
                f.write(col + " " * (width - len(col)) + "|")
            f.write("\n")

            f.write("+")
            for width in widths:
                f.write("-" * width + "+")
            f.write("\n")

        f.seek(0)

        return f.read()



import logging
import os


logger = logging.getLogger("rstcloth.cloth")


class Cloth(object):
    def print_content(self):
        print("\n".join(self._data))  # noqa: T001

    def write(self, filename):
        """

        :param filename:
        :return:
        """
        dirpath = os.path.dirname(filename)
        if os.path.isdir(dirpath) is False:
            try:
                os.makedirs(dirpath)
            except OSError:
                logger.info("{0} exists. ignoring.".format(dirpath))

        with open(filename, "w") as f:
            f.write("\n".join(self._data))
            f.write("\n")

    @property
    def data(self):
        """

        :return:
        """
        return self._data

    @data.setter
    def data(self, value):
        """

        :return:
        """
        raise AttributeError("cannot set the {}.data attribute directly".format(self.__class__.__name__))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rstcloth']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0', 'Pygments>=2.12.0,<3.0.0', 'sphinx>=2,<5']

extras_require = \
{'docs': ['sphinx-rtd-theme==1.0.0',
          'sphinx-tabs==3.2.0',
          'sphinx-charts==0.1.2',
          'sphinx-math-dollar==1.2.0']}

setup_kwargs = {
    'name': 'rstcloth',
    'version': '0.3.2',
    'description': 'A simple Python API for generating RestructuredText.',
    'long_description': "[![Build Status](https://travis-ci.com/thclark/rstcloth.svg?branch=master)](https://travis-ci.com/thclark/rstcloth)\n[![codecov](https://codecov.io/gh/thclark/rstcloth/branch/master/graph/badge.svg)](https://codecov.io/gh/thclark/rstcloth)\n[![PyPI version](https://badge.fury.io/py/rstcloth.svg)](https://badge.fury.io/py/rstcloth)\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/pre-commit/pre-commit)\n[![black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n[![black-girls-code](https://img.shields.io/badge/black%20girls-code-f64279.svg)](https://www.blackgirlscode.com/)\n[![Documentation Status](https://readthedocs.org/projects/rstcloth/badge/?version=latest)](https://rstcloth.readthedocs.io/en/latest/?badge=latest)\n\n\n# RstCloth\n\nreStructuredText is a powerful human-centric markup language that is\nwell defined, flexible, with powerful tools that make writing and\nmaintaining text easy and pleasurable. Humans can edit\nreStructuredText without the aide of complex editing tools, and the\nresulting source is easy to manipulate and process.\n\nAs an alternative and a supplement, RstCloth is a Python API for\nwriting well formed reStructuredText programatically.\n\nFind the [project documentation here](https://rstcloth.readthedocs.io)\n\n\n## Developer notes\n\nRepo is based on [thclark/python-library-template](https://github.com/thclark/python-library-template):\n - black style\n - sphinx docs with some examples and automatic build\n - pre-commit hooks\n - tox tests\n - travis ci + cd\n - code coverage\n\n### Pre-Commit\n\nYou need to install pre-commit to get the hooks working. Do:\n```\npip install pre-commit\npre-commit install\n```\n\nOnce that's done, each time you make a commit, the following checks are made:\n\n- valid github repo and files\n- code style\n- import order\n- PEP8 compliance\n- documentation build\n- branch naming convention\n\nUpon failure, the commit will halt. **Re-running the commit will automatically fix most issues** except:\n\n- The flake8 checks... hopefully over time Black (which fixes most things automatically already) will negate need for it.\n- You'll have to fix documentation yourself prior to a successful commit (there's no auto fix for that!!).\n\nYou can run pre-commit hooks without making a commit, too, like:\n```\npre-commit run black --all-files\n```\nor\n```\n# -v gives verbose output, useful for figuring out why docs won't build\npre-commit run build-docs -v\n```\n\n\n### Contributing\n\n- Please raise an issue on the board (or add your $0.02 to an existing issue) so the maintainers know\nwhat's happening and can advise / steer you.\n\n- Create a fork of rstcloth, undertake your changes on a new branch, (see `.pre-commit-config.yaml` for branch naming conventions). To run tests and make commits,\nyou'll need to do something like:\n```\ngit clone <your_forked_repo_address>    # fetches the repo to your local machine\ncd rstcloth                    # move into the repo directory\npyenv virtualenv 3.6.9 myenv            # Makes a virtual environment for you to install the dev tools into. Use any python >= 3.6\npyend activate myenv                    # Activates the virtual environment so you don't screw up other installations\npip install -r requirements-dev.txt     # Installs the testing and code formatting utilities\npre-commit install                      # Installs the pre-commit code formatting hooks in the git repo\ntox                                     # Runs the tests with coverage. NB you can also just set up pycharm or vscode to run these.\n```\n\n- Adopt a Test Driven Development approach to implementing new features or fixing bugs.\n\n- Ask the `rstcloth` maintainers *where* to make your pull request. We'll create a version branch, according to the\nroadmap, into which you can make your PR. We'll help review the changes and improve the PR.\n\n- Once checks have passed, test coverage of the new code is >=95%, documentation is updated and the Review is passed, we'll merge into the version branch.\n\n- Once all the roadmapped features for that version are done, we'll release.\n\n\n### Release process\n\nThe process for creating a new release is as follows:\n\n1. Check out a branch for the next version, called `vX.Y.Z`\n2. Create a Pull Request into the `master` branch.\n3. Undertake your changes, committing and pushing to branch `vX.Y.Z`\n4. Ensure that documentation is updated to match changes, and increment the changelog. **Pull requests which do not update documentation will be refused.**\n5. Ensure that test coverage is sufficient. **Pull requests that decrease test coverage will be refused.**\n6. Ensure code meets style guidelines (pre-commit scripts and flake8 tests will fail otherwise)\n7. Address Review Comments on the PR\n8. Ensure the version in `setup.py` is correct and matches the branch version.\n9. Merge to master. Successful test, doc build, flake8 and a new version number will automatically create the release on pypi.\n10. Go to code > releases and create a new release on GitHub at the same SHA.\n\n\n## Documents\n\n### Building documents automatically\n\nThe documentation will build automatically in a pre-configured environment when you make a commit.\n\nIn fact, the way pre-commit works, you won't be allowed to make the commit unless the documentation builds,\nthis way we avoid getting broken documentation pushed to the main repository on any commit sha, so we can rely on\nbuilds working.\n\n\n### Building documents manually\n\n**If you did need to build the documentation**\n\nInstall `doxgen`. On a mac, that's `brew install doxygen`; other systems may differ.\n\nInstall sphinx and other requirements for building the docs:\n```\npip install -r docs/requirements.txt\n```\n\nRun the build process:\n```\nsphinx-build -b html docs/source docs/build\n```\n",
    'author': 'Tom Clark',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/thclark/rstcloth',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)
